#ifndef __ALLOC_H__
#define __ALLOC_H__

// Rukia: An allocator must derive from the following class, overload all functions, and provide all default functionality.
//	It may, but does not have to, provide the eact same default template arguments.

#include <new>
#include <string.h>

/*
template < typename AllocType = void*>
class Alloc
{
	public:
	
	// Rukia: Malloc-esq allocation. STL compliant, though does not use it.
	static void* Allocate(size_t amount);
	static void* Reallocate(void* ptr, size_t newsize);
	static void Deallocate(void* ptr);
}
*/

// Rukia: The actual allocators.

// Rukia: Our standard malloc allocator
#include <malloc.h>

template < typename AllocType = void*>
class MAlloc
{
	public:
	
	static void* Allocate(size_t amount) 
		{ return malloc(amount); }

	static void* Reallocate(void* ptr, size_t newsize) 
		{ return realloc(ptr,newsize); }

	static void Deallocate(void* ptr)
		{ free(ptr); }
};

typedef MAlloc<> Mallocator;

#include <string.h>

template < typename AllocType = void*>
class CPPAlloc
{
	public:
	
	static void* Allocate(size_t amount) 
		{ return (void*)(new char[amount]); }

	static void* Reallocate(void* ptr, size_t newsize) 
	{ 
			register void* tmp = (void*)(new char[newsize]); 
			memcpy(tmp,ptr,newsize); 

			delete[] ptr; 
			return tmp; 
	}

	static void Deallocate(void* ptr)
		{ delete[] ptr; }
};

typedef CPPAlloc<> Nallocator;

	// Rukia: Turns out these are a lot more trouble than they are worth for now.
	//	These are C++-esq wrappers. Not STL compliant, but useful.
	//	Will work no matter how the allocator runs, as long as it provides the above functions.


	template <typename T, typename Alloc> 
	static T* New(size_t amount = 1, const T& def = T()) 
	{ 
			T* buffer = (T*)Alloc::Allocate(amount * sizeof(T)); 
			for(register size_t i = 0; i < amount; i++ ) new ((void*)(buffer + i)) T(def);

			return buffer;
	}

	template <typename T, typename Alloc> 
	static void Delete(T* ptr, size_t amount = 1) 
	{ 
		if(ptr == NULL) { return; } 
		
			// Rukia: Roll through all of the elements, destroy each and every one.
			register T* ptr2 = ptr; 
			for(register size_t tmp = 0; tmp < amount; tmp++) 
			{ 
				ptr2->~T(); 
				ptr2++; 
			} 
			Alloc::Deallocate(ptr); 
	}
		
	template <typename T, typename Alloc> 
	static T* More(T* ptr, size_t old_size, size_t new_size, const T& def = T()) 
	{ 
		// Rukia: Woops, someone made a mistake. Don't screw up their work!
		if(new_size == old_size) { return ptr; } 

		// Rukia: Lets make their array smaller 
		else if(new_size < old_size) 
		{ 
			// Rukia: Destroy all the ecess elements
			for(register T* ptr2 = ptr + new_size; ptr2 < (ptr + old_size); ptr2++) 
			{ 
				ptr2->~T(); 
			} 
			Alloc::Reallocate( (void*)ptr, new_size * sizeof(T) ); 
		} 
		// Rukia: They want more, give em more
		else 
		{ 
			Alloc::Reallocate( (void*)ptr,new_size * sizeof(T)); 

			// Rukia: Construct them all, one by one, because C++ doesn't have any array constructor. 
			//	Which really is annoying.
			register T* ptr2 = ptr + old_size; 
			for(register size_t i = 0; i < (new_size - old_size); i++ ) new ((void*)(ptr2 + i)) T(def);
		} 

		return ptr; 
	}

#endif
