#ifndef __HASH_MAP___
#define __HASH_MAP___

// Rukia: Normally, I dislike hashtables. Too memory ineffecient for my tastes.
//	This hashtable, however, is quite delectable.
//	Its called a "cockoo hash"; it allows for all amortized constant time operations.
//	This, of course, is an almost unachievable (and perhaps not wanted) standard.
//	Ours will have worst case O(1) lookups and deletes, and worst case O(log n) insertions.
//	As always with hashtables, no sorted traversal.

// Rukia: While hash_map is not STL standard, it is implimented in 90% of all STL's.
//	Sometimes you want a map, sometimes a hash_map.
//	While our map will put ALL hashtables and maps to shame, this will buy us some time.

// Rukia: NOTE: I am NOT going to make all the function objects. Too much F'n time. I ain't implimenting an STL here.
//	The default values for hash_map will be defined, but the rest will be left for a rainy day. Very. Rainy. Day.

#include "vector.h"
#include "pair.h"
#include "functional"

// Rukia: Conceptually, the hash_map is just a map of hashed values; not terribly different than a normal map.
//	The problem becomes, what kind of hashtable to mapify?
//	Chaining degrades gracefully in performance, but wastes space in the short term.
//	Probing degrades dramatically, saves space in the short term, and wastes it in the long term.
//		Really, no simple hashtable fits the bill; hash_map has to be fast always, and waste little space.

// Rukia: A cockoo hashtable is a probing hashtable which probes in a fixed number of locations, usually 2 or 3.
//	It uses an eviction policy to make sure each and every inserted item gets a slot.
//	This naively makes its performance: lookups and deletes O(k) = O(1), insertions O(n^2)
//	However, by setting an upper limit on how many times the eviction policy can be called, we can improve performance.

// Rukia: We will rehash the table if log(k) items have to be evicted from the table, where k is the size of the table.
//	This makes our worst insertion case for a set M m(log k), for an amortized O(log k) insertion rate.
//	For probing hashing, there can be no better set of garantees.

// Key, Data, Hash, Equal functor, Allocator
//	We will provide our own internal hash algorithm, in addition to Hash.

template < typename K, typename T, typename Hash, typename Equal = equal_to<K>, typename Alloc = MAlloc<K> >
class hash_map
{

};

#endif