#include "amxxmodule.h"
#include "pair"
#include "capsule.h"
#include <string.h>

// Rukia: Pair class is a new addition; it can store TWO of ANYTHING.
//	Using this, you can create a lot of interesting structures, mostly tuples, out of em.

typedef pair<Capsule,Capsule> any_pair;

// Pair:pair_create(Pair:copy = 0, static = 0)
static cell AMX_NATIVE_CALL	pair_create(AMX *amx,cell *params)
{
	any_pair* tmp;

	// Rukia: Copy another pair if necessary.
	if(params[2] == 1) { static any_pair tty; tty.first.clear(); tty.second.clear(); tmp = &tty; }
	else { tmp = New< any_pair, Mallocator >(); }

	if( params[1] != 0)
	{
		any_pair* tty = reinterpret_cast< any_pair * >(params[1]);
		*tmp = *tty;
	}

	return reinterpret_cast<cell>(tmp);
}

// void pair_copy(Pair:designation,Pair:source)
static cell AMX_NATIVE_CALL	pair_copy(AMX *amx,cell *params)
{
	any_pair* dsg = reinterpret_cast<any_pair*>(params[1]);
	any_pair* src = reinterpret_cast<any_pair*>(params[2]);

	*dsg = *src;
	return 1;
}

// void pair_comp(Pair:designation,Pair:source)
static cell AMX_NATIVE_CALL	pair_comp(AMX *amx,cell *params)
{
	any_pair* dsg = reinterpret_cast<any_pair*>(params[1]);
	any_pair* src = reinterpret_cast<any_pair*>(params[2]);

	return (*dsg) == (*src);
}

// void pair_lex(Pair:designation,Pair:source)
static cell AMX_NATIVE_CALL	pair_lex(AMX *amx,cell *params)
{
	any_pair* dsg = reinterpret_cast<any_pair*>(params[1]);
	any_pair* src = reinterpret_cast<any_pair*>(params[2]);

	return (*dsg) < (*src);
}

// Pair:pair_destroy(Pair:me)
static cell AMX_NATIVE_CALL	pair_destroy(AMX *amx,cell *params)
{
	Delete<any_pair, Mallocator >(reinterpret_cast<any_pair*>(params[1]),1);

	return 1;
}

// void pair_clear(Pair:me)
static cell AMX_NATIVE_CALL	pair_clear(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->first.clear();
	reinterpret_cast<any_pair*>(params[1])->second.clear();

	return 1;
}

// bool:pair_empty(Pair:me)
static cell AMX_NATIVE_CALL	pair_empty(AMX *amx,cell *params)
{
	register any_pair* tmp = reinterpret_cast<any_pair*>(params[1]);
	return static_cast<cell>(tmp->first.empty() || tmp->second.empty() );
}

// pair_set_first(Pair:me,Any:newval)
static cell AMX_NATIVE_CALL	pair_set_first(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->first = *reinterpret_cast<Capsule*>(params[2]);
	return 1;
}

// Any:pair_get_first(Pair:me)
static cell AMX_NATIVE_CALL	pair_get_first(AMX *amx,cell *params)
{
	return reinterpret_cast<cell>(&reinterpret_cast<any_pair*>(params[1])->first );
}

// pair_set_first_none(Pair:me)
static cell AMX_NATIVE_CALL	pair_set_first_none(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->first.setnone();
	return 1;
}

// pair_get_first_raw(Pair:me)
static cell AMX_NATIVE_CALL	pair_get_first_raw(AMX *amx,cell *params)
{
	return reinterpret_cast<cell>(reinterpret_cast<any_pair*>(params[1])->first.getData() );
}

// pair_set_first_bool(Pair:me,bool:newval)
static cell AMX_NATIVE_CALL	pair_set_first_bool(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->first.setbool( (params[2] == 1) );
	return 1;
}

// bool:pair_get_first_bool(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_first_bool(AMX *amx,cell *params)
{
	bool* tmp = reinterpret_cast<any_pair*>(params[1])->first.getbool();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve boolean from non-boolean ANY class"); }
		return -1;
	}
	return static_cast<cell>(*tmp);
}

// pair_set_first_int(Pair:me,cell:newval)
static cell AMX_NATIVE_CALL	pair_set_first_int(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->first.setint(params[2]);
	return 1;
}

// cell:pair_get_first_int(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_first_int(AMX *amx,cell *params)
{
	cell* tmp = reinterpret_cast<any_pair*>(params[1])->first.getint();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve integer from non-integer ANY class"); }
		return -1;
	}
	return *tmp;
}

// pair_set_first_ptr(Pair:me,Ptr:newval)
static cell AMX_NATIVE_CALL	pair_set_first_ptr(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->first.setpointer(reinterpret_cast<void*>(params[2]));
	return 1;
}

// Ptr:pair_get_first_ptr(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_first_ptr(AMX *amx,cell *params)
{
	void* tmp = reinterpret_cast<any_pair*>(params[1])->first.getpointer();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve pointer from non-pointer ANY class"); }
		return -1;
	}
	return reinterpret_cast<cell>(tmp);
}

// pair_set_first_float(Pair:me,Float:newval)
static cell AMX_NATIVE_CALL	pair_set_first_float(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->first.setfloat(amx_ctof(params[2]));
	return 1;
}

// Float:pair_get_first_float(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_first_float(AMX *amx,cell *params)
{
	REAL* tmp = reinterpret_cast<any_pair*>(params[1])->first.getfloat();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve float from non-float ANY class"); }
		return -1;
	}
	return amx_ftoc(*tmp);
}

// pair_set_first_vector(Pair:me,Float:newval[3])
static cell AMX_NATIVE_CALL	pair_set_first_vector(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	HLVector tmp(amx_ctof(ptr[0]),amx_ctof(ptr[1]),amx_ctof(ptr[2]) );

	reinterpret_cast<any_pair*>(params[1])->first.setvector(tmp);
	return 1;
}

// pair_get_first_vector(Pair:me,Float:vec[3],no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_first_vector(AMX *amx,cell *params)
{
	HLVector* tmp = reinterpret_cast<any_pair*>(params[1])->first.getvector();
	if( (tmp == NULL) )
	{
		if(params[3] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve vector from non-vector ANY class"); }
		return -1;
	}
	cell* vec = MF_GetAmxAddr(amx,params[2]);

	vec[0] = amx_ftoc(tmp->x);
	vec[1] = amx_ftoc(tmp->y);
	vec[2] = amx_ftoc(tmp->z);

	return 1;
}

// pair_set_first_array(Pair:me,newval[],len)
static cell AMX_NATIVE_CALL	pair_set_first_array(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	vector<cell> tmp(ptr,params[3]);

	reinterpret_cast<any_pair*>(params[1])->first.setarray(tmp);
	return 1;
}


// pair_get_first_array(Pair:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_first_array(AMX *amx,cell *params)
{
	register vector<cell>* tmp = reinterpret_cast<any_pair*>(params[1])->first.getarray();
	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve array from non-array ANY class"); }
		return -1;
	}
	register cell* arr = MF_GetAmxAddr(amx,params[2]);
	register size_t len = (static_cast<ucell>(params[3]) < tmp->size()) ? (params[3]) : (tmp->size());
	for(register size_t i = 0; i < len; i++)
	{
		arr[i] = (*tmp)[i];
	}

	return len;
}

// pair_set_first_string(Pair:me,newval[])
static cell AMX_NATIVE_CALL	pair_set_first_string(AMX *amx,cell *params)
{
	int iLen;
	char* ptr = MF_GetAmxString(amx,params[2],0,&iLen);
	String tmp(ptr,iLen);

	reinterpret_cast<any_pair*>(params[1])->first.setstring(tmp);
	return 1;
}


// pair_get_first_string(Pair:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_first_string(AMX *amx,cell *params)
{
	register String* tmp = reinterpret_cast<any_pair*>(params[1])->first.getstring();
	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve string from non-string ANY class"); }
		return -1;
	}

	return MF_SetAmxString(amx,params[2],tmp->c_str(),params[3]);
}

// pair_set_second(Pair:me,Any:newval)
static cell AMX_NATIVE_CALL	pair_set_second(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->second = *reinterpret_cast<Capsule*>(params[2]);
	return 1;
}

// Any:pair_get_second(Pair:me)
static cell AMX_NATIVE_CALL	pair_get_second(AMX *amx,cell *params)
{
	return reinterpret_cast<cell>(&reinterpret_cast<any_pair*>(params[1])->second );
}

// pair_set_second_none(Pair:me)
static cell AMX_NATIVE_CALL	pair_set_second_none(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->second.setnone();
	return 1;
}

// pair_get_second_raw(Pair:me)
static cell AMX_NATIVE_CALL	pair_get_second_raw(AMX *amx,cell *params)
{
	return reinterpret_cast<cell>(reinterpret_cast<any_pair*>(params[1])->second.getData() );
}

// pair_set_second_bool(Pair:me,bool:newval)
static cell AMX_NATIVE_CALL	pair_set_second_bool(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->second.setbool((params[2] == 1));
	return 1;
}

// bool:pair_get_second_bool(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_second_bool(AMX *amx,cell *params)
{
	bool* tmp = reinterpret_cast<any_pair*>(params[1])->second.getbool();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve boolean from non-boolean ANY class"); }
		return -1;
	}
	return static_cast<cell>(*tmp);
}

// pair_set_second_int(Pair:me,cell:newval)
static cell AMX_NATIVE_CALL	pair_set_second_int(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->second.setint(params[2]);
	return 1;
}

// cell:pair_get_second_int(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_second_int(AMX *amx,cell *params)
{
	cell* tmp = reinterpret_cast<any_pair*>(params[1])->second.getint();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve integer from non-integer ANY class"); }
		return -1;
	}
	return *tmp;
}

// pair_set_second_ptr(Pair:me,Ptr:newval)
static cell AMX_NATIVE_CALL	pair_set_second_ptr(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->second.setpointer(reinterpret_cast<void*>(params[2]));
	return 1;
}

// Ptr:pair_get_second_ptr(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_second_ptr(AMX *amx,cell *params)
{
	void* tmp = reinterpret_cast<any_pair*>(params[1])->second.getpointer();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve pointer from non-pointer ANY class"); }
		return -1;
	}
	return reinterpret_cast<cell>(tmp);
}

// pair_set_second_float(Pair:me,Float:newval)
static cell AMX_NATIVE_CALL	pair_set_second_float(AMX *amx,cell *params)
{
	reinterpret_cast<any_pair*>(params[1])->second.setfloat(amx_ctof(params[2]));
	return 1;
}

// Float:pair_get_second_float(Pair:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_second_float(AMX *amx,cell *params)
{
	REAL* tmp = reinterpret_cast<any_pair*>(params[1])->second.getfloat();
	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve float from non-float ANY class"); }
		return -1;
	}
	return amx_ftoc(*tmp);
}

// pair_set_second_vector(Pair:me,Float:newval[3])
static cell AMX_NATIVE_CALL	pair_set_second_vector(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	HLVector tmp(amx_ctof(ptr[0]),amx_ctof(ptr[1]),amx_ctof(ptr[2]) );

	reinterpret_cast<any_pair*>(params[1])->second.setvector(tmp);
	return 1;
}

// pair_get_second_vector(Pair:me,Float:vec[3],no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_second_vector(AMX *amx,cell *params)
{
	HLVector* tmp = reinterpret_cast<any_pair*>(params[1])->second.getvector();
	if( (tmp == NULL) )
	{
		if(params[3] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve vector from non-vector ANY class"); }
		return -1;
	}
	cell* vec = MF_GetAmxAddr(amx,params[2]);

	vec[0] = amx_ftoc(tmp->x);
	vec[1] = amx_ftoc(tmp->y);
	vec[2] = amx_ftoc(tmp->z);

	return 1;
}

// pair_set_second_array(Pair:me,newval[],len)
static cell AMX_NATIVE_CALL	pair_set_second_array(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	vector<cell> tmp(ptr,params[3]);

	reinterpret_cast<any_pair*>(params[1])->second.setarray(tmp);
	return 1;
}


// pair_get_second_array(Pair:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_second_array(AMX *amx,cell *params)
{
	register vector<cell>* tmp = reinterpret_cast<any_pair*>(params[1])->second.getarray();
	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve array from non-array ANY class"); }
		return -1;
	}
	register cell* arr = MF_GetAmxAddr(amx,params[2]);
	register size_t len = (static_cast<ucell>(params[3]) < tmp->size()) ? (params[3]) : (tmp->size());
	for(register size_t i = 0; i < len; i++)
	{
		arr[i] = (*tmp)[i];
	}

	return len;
}

// pair_set_second_string(Pair:me,newval[])
static cell AMX_NATIVE_CALL	pair_set_second_string(AMX *amx,cell *params)
{
	int iLen;
	char* ptr = MF_GetAmxString(amx,params[2],0,&iLen);
	String tmp(ptr,iLen);

	reinterpret_cast<any_pair*>(params[1])->second.setstring(tmp);
	return 1;
}


// pair_get_second_string(Pair:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	pair_get_second_string(AMX *amx,cell *params)
{
	register String* tmp = reinterpret_cast<any_pair*>(params[1])->second.getstring();
	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve string from non-string ANY class"); }
		return -1;
	}

	return MF_SetAmxString(amx,params[2],tmp->c_str(),params[3]);
}

AMX_NATIVE_INFO pair_Natives[] =
{
	{"pair_create",		pair_create},
	{"pair_destroy",		pair_destroy},

	{"pair_clear",		pair_clear},
	{"pair_copy",		pair_copy},

	{"pair_comp",		pair_comp},
	{"pair_lex",		pair_lex},

	{"pair_empty",		pair_empty},

	{"pair_set_first",			pair_set_first},
	{"pair_get_first",			pair_get_first},

	{"pair_set_first_none",		pair_set_first_none},
	{"pair_get_first_raw",		pair_get_first_raw},

	{"pair_set_first_bool",		pair_set_first_bool},
	{"pair_get_first_bool",		pair_get_first_bool},

	{"pair_set_first_int",		pair_set_first_int},
	{"pair_get_first_int",		pair_get_first_int},

	{"pair_set_first_ptr",		pair_set_first_ptr},
	{"pair_get_first_ptr",		pair_get_first_ptr},

	{"pair_set_first_float",		pair_set_first_float},
	{"pair_get_first_float",		pair_get_first_float},

	{"pair_set_first_vector",		pair_set_first_vector},
	{"pair_get_first_vector",		pair_get_first_vector},

	{"pair_set_first_array",		pair_set_first_array},
	{"pair_get_first_array",		pair_get_first_array},

	{"pair_set_first_string",		pair_set_first_string},
	{"pair_get_first_string",		pair_get_first_string},

//	{"pair_set_first_custom",		pair_set_first_custom},
//	{"pair_get_first_custom",		pair_get_first_custom},

	{"pair_set_second",			pair_set_second},
	{"pair_get_second",			pair_get_second},

	{"pair_set_second_none",		pair_set_second_none},
	{"pair_get_second_raw",			pair_get_second_raw},

	{"pair_set_second_bool",		pair_set_second_bool},
	{"pair_get_second_bool",		pair_get_second_bool},

	{"pair_set_second_int",			pair_set_second_int},
	{"pair_get_second_int",			pair_get_second_int},

	{"pair_set_second_ptr",			pair_set_second_ptr},
	{"pair_get_second_ptr",			pair_get_second_ptr},

	{"pair_set_second_float",		pair_set_second_float},
	{"pair_get_second_float",		pair_get_second_float},

	{"pair_set_second_vector",		pair_set_second_vector},
	{"pair_get_second_vector",		pair_get_second_vector},

	{"pair_set_second_array",		pair_set_second_array},
	{"pair_get_second_array",		pair_get_second_array},

	{"pair_set_second_string",		pair_set_second_string},
	{"pair_get_second_string",		pair_get_second_string},

//	{"pair_set_second_custom",		pair_set_second_custom},
//	{"pair_get_second_custom",		pair_get_second_custom},

	{ NULL,				NULL}
};