#ifndef __PAWN_ITER__
#define __PAWN_ITER__

#include "alloc.h"
#include "capsule.h"
#include <string.h>

#define MUTABLE_THIS (pawnIterator*)this

template <typename Alloc = MAlloc<char> >
class pawnIterator
{
public:
enum
{
	// (void) (NULL,NULL)
	ITERATOR_OP_CONSTRUCT = 0,

	// (NULL) (this,NULL)
	ITERATOR_OP_COPY,
	//(void) (this,NULL)
	ITERATOR_OP_DECONSTRUCT,

	// (void) (this,amount)
	ITERATOR_OP_PREFIX_ADD,
	// (void) (this,amount)
	ITERATOR_OP_PREFIX_SUB,

	// (void*) (this,amount)
	//ITERATOR_OP_POSTFIX_ADD,
	// (void*) (this,amount)
	//ITERATOR_OP_POSTFIX_SUB,

	// (bool) (this,other)
	ITERATOR_OP_EQUALTO,
	// (bool) (this,other)
	ITERATOR_OP_LESSTHAN,

	// (Capsule*) (this,NULL)
	ITERATOR_OP_DEREFERENCE,
	// (Capsule*) (this,NULL)
	ITERATOR_OP_KEY
};

public:
	// Rukia: Function for our virtual calls. Takes *this, and the op code we are calling, and a tag.
	//	Returns something we have to cast, and probably delete.
	void* (*iterfunc)(pawnIterator*,size_t,void*);

	// Rukia: The actual iterator pointer. Needs the above function pointer to cast it.j
	void* iteratorptr;

	pawnIterator( void* (*Func)(pawnIterator*,size_t,void*) ) : 
		iterfunc(Func), 
		iteratorptr(Func(NULL,ITERATOR_OP_CONSTRUCT,NULL) 
	)
	{ }

	~pawnIterator()
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_DECONSTRUCT,NULL); }

	template <typename T>
	void set(T iterat)
		{ *(T*)(iteratorptr) = iterat; }

	pawnIterator& operator=(const pawnIterator* copy)
		{  iterfunc(MUTABLE_THIS,ITERATOR_OP_COPY,(void*)(&copy)); return *this; }

	pawnIterator& operator++()
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_PREFIX_ADD,(void*)(1)); return *this; }

	pawnIterator& operator--()
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_PREFIX_SUB,(void*)(1)); return *this; }

	pawnIterator& operator+(size_t amount)
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_PREFIX_ADD,(void*)(amount)); return *this; }

	pawnIterator& operator-(size_t amount)
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_PREFIX_SUB,(void*)(amount)); return *this; }

	pawnIterator& operator+=(size_t amount)
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_PREFIX_ADD,(void*)(amount)); return *this; }

	pawnIterator& operator-=(size_t amount)
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_PREFIX_SUB,(void*)(amount)); return *this; }

	// Rukia: No postfix. Too expensive, and too difficult to make work correctly.
	//	We use postfix for loops; doing that here will generate a probably unused heap allocated iterator.
	/*
	pawnIterator& operator++(int)
		{ return iterfunc(MUTABLE_THIS,ITERATOR_OP_POSTFIX_ADD,(void*)(1)); }

	pawnIterator& operator--(int)
		{ iterfunc(MUTABLE_THIS,ITERATOR_OP_POSTFIX_SUB,(void*)(1)); }
		*/

	bool operator<(const pawnIterator& other)
		{ return iterfunc(MUTABLE_THIS,ITERATOR_OP_LESSTHAN,(void*)(&other)) == NULL; }

	bool operator==(const pawnIterator& other)
		{ return iterfunc(MUTABLE_THIS,ITERATOR_OP_EQUALTO,(void*)(&other)) == NULL; }

	Capsule& operator*()
	{
		return *reinterpret_cast<Capsule*>(iterfunc(MUTABLE_THIS,ITERATOR_OP_DEREFERENCE,(void*)NULL) );
	}

	Capsule& iterAt()
	{
		return *reinterpret_cast<Capsule*>(iterfunc(MUTABLE_THIS,ITERATOR_OP_KEY,(void*)NULL) );
	}
};

typedef pawnIterator<> PIterator;
typedef PIterator any_iterator;

#endif
