#ifndef __QUEUE___
#define __QUEUE___
// Rukia: Good, another quick one! queues are based on deques, and the only reason I made deques was for queues!

#include "alloc.h"
#include "deque"

template <typename T, typename Sequence = deque<T> >
class queue
{
public: CONTAINER_TYPEDEF;
private: 
	Sequence internal_;
public:
	queue() : internal_()
	{ }
	queue(const queue& copy) : internal_(copy.internal_)
	{ }

	queue& operator=(const queue& copy)
		{ internal_ = copy.internal_; return *this; }

	bool operator==(const queue& other)
		{ return (internal_ == other.internal_); }

	bool operator<(const queue& other)
		{ return (internal_ < other.internal_); }

	void clear()
		{ internal_.clear(); }

	bool empty() const
		{ return internal_.empty(); }

	size_t size() const
		{ return internal_.size(); }

	reference front()
		{ return internal_.front(); }

	const_reference front() const
		{ return internal_.front(); }

	reference back()
		{ return internal_.back(); }

	const_reference back() const
		{ return internal_.back(); }

	void push(reference tmp)
		{ internal_.push_back(tmp); }

	void pop()
		{ internal_.pop_front(); }

	reference peek()
		{ return internal_.front(); }
	
	const_reference peek() const
		{ return internal_.front(); }
};

// Rukia: Compare size first; if the sizes are not equal, they cannot be equal. This is a huge performance boost.
template <typename T, typename Alloc>
bool operator== (const queue<T,Alloc>& a, const queue<T,Alloc>& b)
{ 
	return ( (a.size() == b.size()) && (a == b) );
}

template <typename T, typename Alloc>
bool operator< (const queue<T,Alloc>& a, const queue<T,Alloc>& b)
{
	return ( (a < b) );
}

template < typename T, typename Alloc> inline bool operator != (const queue<T,Alloc>& a, const queue<T,Alloc>& b)
	{ return !(a == b); }

template < typename T, typename Alloc> inline bool operator > (const queue<T,Alloc>& a, const queue<T,Alloc>& b)
	{ return (b < a); }

template < typename T, typename Alloc> inline bool operator >= (const queue<T,Alloc>& a, const queue<T,Alloc>& b)
	{ return !(a < b); }

template < typename T, typename Alloc> inline bool operator <= (const queue<T,Alloc>& a, const queue<T,Alloc>& b)
	{ return !(b < a); }


#endif
