#include "amxxmodule.h"
#include "queue"
#include "capsule.h"

// Rukia: Queues first; no iterators to deal with, simple interface.

typedef queue<Capsule> any_queue;

// Queue:queue_create(Queue:copy = NULL)
static cell AMX_NATIVE_CALL	queue_create(AMX *amx,cell *params)
{
	any_queue* tmp;

	// Rukia: Copy another queue if necessary.
	if(params[2] == 1) { static any_queue tty; tmp = &tty; }
	else { tmp = New<any_queue, Mallocator >(); }
	if(params[1] != 0)
	{
		any_queue * tty = reinterpret_cast<any_queue * >(params[1]);
		*tmp = *tty;
	}

	return reinterpret_cast<cell>(tmp);
}

// void queue_copy(Queue:designation,Queue:source)
static cell AMX_NATIVE_CALL	queue_copy(AMX *amx,cell *params)
{
	any_queue* dsg = reinterpret_cast<any_queue*>(params[1]);
	any_queue* src = reinterpret_cast<any_queue*>(params[2]);

	*dsg = *src;
	return 1;
}

// void bool:queue_comp(Queue:designation,Queue:source)
static cell AMX_NATIVE_CALL	queue_comp(AMX *amx,cell *params)
{
	any_queue* dsg = reinterpret_cast<any_queue*>(params[1]);
	any_queue* src = reinterpret_cast<any_queue*>(params[2]);

	return (*dsg) == (*src);
}

// void bool:queue_lex(Queue:designation,Queue:source)
static cell AMX_NATIVE_CALL	queue_lex(AMX *amx,cell *params)
{
	any_queue* dsg = reinterpret_cast<any_queue*>(params[1]);
	any_queue* src = reinterpret_cast<any_queue*>(params[2]);

	return (*dsg) < (*src);
}

// Rukia: Not technically part of ANSI STL queues; it matches our standard API though.
// void queue_clear(Queue:me)
static cell AMX_NATIVE_CALL	queue_clear(AMX *amx,cell *params)
{
	reinterpret_cast<any_queue*>(params[1])->clear();

	return 1;
}

// Queue:queue_destroy(Queue:me)
static cell AMX_NATIVE_CALL	queue_destroy(AMX *amx,cell *params)
{
	Delete<any_queue, Mallocator >(reinterpret_cast<any_queue*>(params[1]),1);

	return 1;
}

// bool:queue_empty(Queue:me)
static cell AMX_NATIVE_CALL	queue_empty(AMX *amx,cell *params)
{
	return static_cast<cell>(reinterpret_cast<any_queue*>(params[1])->empty() );
}

// Rukia: Unfortunately, to get this out, I have to give up finishing the allocator for a while.
//	Returns the amount of space wasted but not the exact total memory used.

// cell queue_memory(Queue:me)
static cell AMX_NATIVE_CALL	queue_memory(AMX *amx,cell *params)
{
	return static_cast<cell>(reinterpret_cast<any_queue*>(params[1])->size() * sizeof(Capsule) );
}

// cell queue_size(Queue:me)
static cell AMX_NATIVE_CALL	queue_size(AMX *amx,cell *params)
{
	return static_cast<cell>(reinterpret_cast<any_queue*>(params[1])->size() );
}

// queue_pop(Queue:me)
static cell AMX_NATIVE_CALL	queue_pop(AMX *amx,cell *params)
{
	reinterpret_cast<any_queue*>(params[1])->pop();
	return 1;
}

// queue_push(Queue:me,Any:add)
static cell AMX_NATIVE_CALL	queue_push(AMX *amx,cell *params)
{
	reinterpret_cast<any_queue*>(params[1])->
		push(*reinterpret_cast<Capsule*>(params[2]) );
	return 1;
}

// queue_push_none(Queue:me)
static cell AMX_NATIVE_CALL	queue_push_none(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setnone();
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// queue_push_bool(Queue:me,bool:newval)
static cell AMX_NATIVE_CALL	queue_push_bool(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setbool((params[2] == 1));
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// queue_push_int(Queue:me,cell:newval)
static cell AMX_NATIVE_CALL	queue_push_int(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setint(params[2]);
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// queue_push_ptr(Queue:me,Ptr:newval)
static cell AMX_NATIVE_CALL	queue_push_ptr(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setpointer(reinterpret_cast<void*>(params[2]));
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// queue_push_float(Queue:me,Float:newval)
static cell AMX_NATIVE_CALL	queue_push_float(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setfloat(amx_ctof(params[2]));
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// queue_push_queue(Queue:me,Float:newval[3])
static cell AMX_NATIVE_CALL	queue_push_queue(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	HLVector tty(amx_ctof(ptr[0]),amx_ctof(ptr[1]),amx_ctof(ptr[2]));

	Capsule tmp; tmp.setvector(tty);
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// queue_push_array(Queue:me,newval[],len)
static cell AMX_NATIVE_CALL	queue_push_array(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	vector<cell> tty(ptr,params[3]);

	Capsule tmp; tmp.setarray(tty);
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// queue_push_string(Queue:me,newval[])
static cell AMX_NATIVE_CALL	queue_push_string(AMX *amx,cell *params)
{
	int iLen;
	char* ptr = MF_GetAmxString(amx,params[2],0,&iLen);
	String tty(ptr,iLen);

	Capsule tmp; tmp.setstring(tty);
	reinterpret_cast<any_queue*>(params[1])->push(tmp);
	return 1;
}

// Any:queue_peek(Queue:me)
static cell AMX_NATIVE_CALL	queue_peek(AMX *amx,cell *params)
{
	Capsule& tmp = reinterpret_cast<any_queue*>(params[1])->peek();
	return reinterpret_cast<cell>(&tmp);
}

// queue_peek_raw(Queue:me)
static cell AMX_NATIVE_CALL	queue_peek_raw(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	
	return reinterpret_cast<cell>(tmp_queue->peek().getData() );
}

// bool:queue_peek_bool(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_peek_bool(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	bool* tmp = tmp_queue->peek().getbool();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve boolean from non-boolean ANY class"); }
		return -1;
	}
	return static_cast<cell>(*tmp);
}

// cell:queue_peek_int(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_peek_int(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	cell* tmp = tmp_queue->peek().getint();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve integer from non-integer ANY class"); }
		return -1;
	}
	return *tmp;
}

// Ptr:queue_peek_ptr(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_peek_ptr(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	void* tmp = tmp_queue->peek().getpointer();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve pointer from non-pointer ANY class"); }
		return -1;
	}
	return reinterpret_cast<cell>(tmp);
}

// Float:queue_peek_float(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_peek_float(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	REAL* tmp = tmp_queue->peek().getfloat();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve float from non-float ANY class"); }
		return -1;
	}
	return amx_ftoc(*tmp);
}

// queue_peek_queue(Queue:me,Float:vec[3],no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_peek_queue(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	HLVector* tmp = tmp_queue->peek().getvector();

	if( (tmp == NULL) )
	{
		if(params[3] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve queue from non-queue ANY class"); }
		return -1;
	}
	cell* vec = MF_GetAmxAddr(amx,params[2]);

	vec[0] = amx_ftoc(tmp->x);
	vec[1] = amx_ftoc(tmp->y);
	vec[2] = amx_ftoc(tmp->z);

	return 1;
}

// queue_peek_array(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_peek_array(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	register vector<cell>* tmp = tmp_queue->peek().getarray();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve array from non-array ANY class"); }
		return -1;
	}
	register cell* arr = MF_GetAmxAddr(amx,params[2]);
	register size_t len = (static_cast<ucell>(params[3]) < tmp->size()) ? (params[3]) : (tmp->size());
	for(register size_t i = 0; i < len; i++)
	{
		arr[i] = (*tmp)[i];
	}

	return len;
}

// queue_peek_string(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_peek_string(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	register String* tmp = tmp_queue->peek().getstring();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve string from non-string ANY class"); }
		return -1;
	}

	return MF_SetAmxString(amx,params[2],tmp->c_str(),params[3]);
}

// Any:queue_front(Queue:me)
static cell AMX_NATIVE_CALL	queue_front(AMX *amx,cell *params)
{
	return reinterpret_cast<cell>(&reinterpret_cast<any_queue*>(params[1])->front() );
}

// queue_front_raw(Queue:me)
static cell AMX_NATIVE_CALL	queue_front_raw(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	
	return reinterpret_cast<cell>(tmp_queue->front().getData() );
}

// bool:queue_front_bool(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_front_bool(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	bool* tmp = tmp_queue->front().getbool();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve boolean from non-boolean ANY class"); }
		return -1;
	}
	return static_cast<cell>(*tmp);
}

// cell:queue_front_int(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_front_int(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	cell* tmp = tmp_queue->front().getint();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve integer from non-integer ANY class"); }
		return -1;
	}
	return *tmp;
}

// Ptr:queue_front_ptr(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_front_ptr(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	void* tmp = tmp_queue->front().getpointer();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve pointer from non-pointer ANY class"); }
		return -1;
	}
	return reinterpret_cast<cell>(tmp);
}

// Float:queue_front_float(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_front_float(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	REAL* tmp = tmp_queue->front().getfloat();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve float from non-float ANY class"); }
		return -1;
	}
	return amx_ftoc(*tmp);
}

// queue_front_queue(Queue:me,Float:vec[3],no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_front_queue(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	HLVector* tmp = tmp_queue->front().getvector();

	if( (tmp == NULL) )
	{
		if(params[3] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve queue from non-queue ANY class"); }
		return -1;
	}
	cell* vec = MF_GetAmxAddr(amx,params[2]);

	vec[0] = amx_ftoc(tmp->x);
	vec[1] = amx_ftoc(tmp->y);
	vec[2] = amx_ftoc(tmp->z);

	return 1;
}

// queue_front_array(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_front_array(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	register vector<cell>* tmp = tmp_queue->front().getarray();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve array from non-array ANY class"); }
		return -1;
	}
	register cell* arr = MF_GetAmxAddr(amx,params[2]);
	register size_t len = (static_cast<ucell>(params[3]) < tmp->size()) ? (params[3]) : (tmp->size());
	for(register size_t i = 0; i < len; i++)
	{
		arr[i] = (*tmp)[i];
	}

	return len;
}

// queue_front_string(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_front_string(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	register String* tmp = tmp_queue->front().getstring();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve string from non-string ANY class"); }
		return -1;
	}

	return MF_SetAmxString(amx,params[2],tmp->c_str(),params[3]);
}

// Any:queue_back(Queue:me)
static cell AMX_NATIVE_CALL	queue_back(AMX *amx,cell *params)
{
	return reinterpret_cast<cell>(&reinterpret_cast<any_queue*>(params[1])->back() );
}

// queue_back_raw(Queue:me)
static cell AMX_NATIVE_CALL	queue_back_raw(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	
	return reinterpret_cast<cell>(tmp_queue->back().getData() );
}

// bool:queue_back_bool(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_back_bool(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	bool* tmp = tmp_queue->back().getbool();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve boolean from non-boolean ANY class"); }
		return -1;
	}
	return static_cast<cell>(*tmp);
}

// cell:queue_back_int(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_back_int(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	cell* tmp = tmp_queue->back().getint();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve integer from non-integer ANY class"); }
		return -1;
	}
	return *tmp;
}

// Ptr:queue_back_ptr(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_back_ptr(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	void* tmp = tmp_queue->back().getpointer();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve pointer from non-pointer ANY class"); }
		return -1;
	}
	return reinterpret_cast<cell>(tmp);
}

// Float:queue_back_float(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_back_float(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	REAL* tmp = tmp_queue->back().getfloat();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve float from non-float ANY class"); }
		return -1;
	}
	return amx_ftoc(*tmp);
}

// queue_back_queue(Queue:me,Float:vec[3],no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_back_queue(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	HLVector* tmp = tmp_queue->back().getvector();

	if( (tmp == NULL) )
	{
		if(params[3] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve queue from non-queue ANY class"); }
		return -1;
	}
	cell* vec = MF_GetAmxAddr(amx,params[2]);

	vec[0] = amx_ftoc(tmp->x);
	vec[1] = amx_ftoc(tmp->y);
	vec[2] = amx_ftoc(tmp->z);

	return 1;
}

// queue_back_array(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_back_array(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	register vector<cell>* tmp = tmp_queue->back().getarray();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve array from non-array ANY class"); }
		return -1;
	}
	register cell* arr = MF_GetAmxAddr(amx,params[2]);
	register size_t len = (static_cast<ucell>(params[3]) < tmp->size()) ? (params[3]) : (tmp->size());
	for(register size_t i = 0; i < len; i++)
	{
		arr[i] = (*tmp)[i];
	}

	return len;
}

// queue_back_string(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	queue_back_string(AMX *amx,cell *params)
{
	any_queue* tmp_queue = reinterpret_cast<any_queue*>(params[1]);
	register String* tmp = tmp_queue->back().getstring();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve string from non-string ANY class"); }
		return -1;
	}

	return MF_SetAmxString(amx,params[2],tmp->c_str(),params[3]);
}

AMX_NATIVE_INFO queue_Natives[] =
{
	{"queue_create",		queue_create},
	{"queue_destroy",		queue_destroy},

	{"queue_clear",		queue_clear},
	{"queue_copy",		queue_copy},

	{"queue_comp",		queue_comp},
	{"queue_lex",		queue_lex},

	{"queue_empty",		queue_empty},
	{"queue_size",		queue_size},
	{"queue_memory",	queue_memory},

	{"queue_pop",		queue_pop},

	{"queue_peek",				queue_peek},
	{"queue_peek_raw",			queue_peek_raw},
	{"queue_peek_bool",			queue_peek_bool},
	{"queue_peek_int",			queue_peek_int},
	{"queue_peek_float",		queue_peek_float},
	{"queue_peek_ptr",			queue_peek_ptr},
	{"queue_peek_queue",		queue_peek_queue},
	{"queue_peek_array",		queue_peek_array},
	{"queue_peek_string",		queue_peek_string},

	{"queue_push",				queue_push},
	{"queue_push_none",			queue_push_none},
	{"queue_push_bool",			queue_push_bool},
	{"queue_push_int",			queue_push_int},
	{"queue_push_float",		queue_push_float},
	{"queue_push_ptr",			queue_push_ptr},
	{"queue_push_queue",		queue_push_queue},
	{"queue_push_array",		queue_push_array},
	{"queue_push_string",		queue_push_string},

	{"queue_front",				queue_front},
	{"queue_front_raw",			queue_front_raw},
	{"queue_front_bool",		queue_front_bool},
	{"queue_front_int",			queue_front_int},
	{"queue_front_float",		queue_front_float},
	{"queue_front_ptr",			queue_front_ptr},
	{"queue_front_queue",		queue_front_queue},
	{"queue_front_array",		queue_front_array},
	{"queue_front_string",		queue_front_string},

	{"queue_back",				queue_back},
	{"queue_back_raw",			queue_back_raw},
	{"queue_back_bool",			queue_back_bool},
	{"queue_back_int",			queue_back_int},
	{"queue_back_float",		queue_back_float},
	{"queue_back_ptr",			queue_back_ptr},
	{"queue_back_queue",		queue_back_queue},
	{"queue_back_array",		queue_back_array},
	{"queue_back_string",		queue_back_string},

	{ NULL,				NULL}
};