#include "amxxmodule.h"
#include "stack"
#include "capsule.h"

// Rukia: Queues first; no iterators to deal with, simple interface.

typedef stack<Capsule> any_stack;

// Queue:stack_create(Queue:copy = NULL)
static cell AMX_NATIVE_CALL	stack_create(AMX *amx,cell *params)
{
	any_stack* tmp;

	// Rukia: Copy another stack if necessary.
	if(params[2] == 1) { static any_stack tty; tmp = &tty; }
	else { tmp = New<any_stack, Mallocator >(); }
	if(params[1] != 0)
	{
		any_stack * tty = reinterpret_cast<any_stack * >(params[1]);
		*tmp = *tty;
	}

	return reinterpret_cast<cell>(tmp);
}

// void stack_copy(Queue:designation,Queue:source)
static cell AMX_NATIVE_CALL	stack_copy(AMX *amx,cell *params)
{
	any_stack* dsg = reinterpret_cast<any_stack*>(params[1]);
	any_stack* src = reinterpret_cast<any_stack*>(params[2]);

	*dsg = *src;
	return 1;
}

// void bool:stack_comp(Queue:designation,Queue:source)
static cell AMX_NATIVE_CALL	stack_comp(AMX *amx,cell *params)
{
	any_stack* dsg = reinterpret_cast<any_stack*>(params[1]);
	any_stack* src = reinterpret_cast<any_stack*>(params[2]);

	return (*dsg) == (*src);
}

// void bool:stack_lex(Queue:designation,Queue:source)
static cell AMX_NATIVE_CALL	stack_lex(AMX *amx,cell *params)
{
	any_stack* dsg = reinterpret_cast<any_stack*>(params[1]);
	any_stack* src = reinterpret_cast<any_stack*>(params[2]);

	return (*dsg) < (*src);
}

// Rukia: Not technically part of ANSI STL stacks; it matches our standard API though.
// void stack_clear(Queue:me)
static cell AMX_NATIVE_CALL	stack_clear(AMX *amx,cell *params)
{
	reinterpret_cast<any_stack*>(params[1])->clear();

	return 1;
}

// Queue:stack_destroy(Queue:me)
static cell AMX_NATIVE_CALL	stack_destroy(AMX *amx,cell *params)
{
	Delete<any_stack, Mallocator >(reinterpret_cast<any_stack*>(params[1]),1);

	return 1;
}

// bool:stack_empty(Queue:me)
static cell AMX_NATIVE_CALL	stack_empty(AMX *amx,cell *params)
{
	return static_cast<cell>(reinterpret_cast<any_stack*>(params[1])->empty() );
}

// Rukia: Unfortunately, to get this out, I have to give up finishing the allocator for a while.
//	Returns the amount of space wasted but not the exact total memory used.

// cell stack_memory(Queue:me)
static cell AMX_NATIVE_CALL	stack_memory(AMX *amx,cell *params)
{
	return static_cast<cell>(reinterpret_cast<any_stack*>(params[1])->size() * sizeof(Capsule) );
}

// cell stack_size(Queue:me)
static cell AMX_NATIVE_CALL	stack_size(AMX *amx,cell *params)
{
	return static_cast<cell>(reinterpret_cast<any_stack*>(params[1])->size() );
}

// stack_pop(Queue:me)
static cell AMX_NATIVE_CALL	stack_pop(AMX *amx,cell *params)
{
	reinterpret_cast<any_stack*>(params[1])->pop();
	return 1;
}

// stack_push(Queue:me,Any:add)
static cell AMX_NATIVE_CALL	stack_push(AMX *amx,cell *params)
{
	reinterpret_cast<any_stack*>(params[1])->
		push(*reinterpret_cast<Capsule*>(params[2]) );
	return 1;
}

// stack_push_none(Queue:me)
static cell AMX_NATIVE_CALL	stack_push_none(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setnone();
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// stack_push_bool(Queue:me,bool:newval)
static cell AMX_NATIVE_CALL	stack_push_bool(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setbool((params[2] == 1));
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// stack_push_int(Queue:me,cell:newval)
static cell AMX_NATIVE_CALL	stack_push_int(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setint(params[2]);
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// stack_push_ptr(Queue:me,Ptr:newval)
static cell AMX_NATIVE_CALL	stack_push_ptr(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setpointer(reinterpret_cast<void*>(params[2]));
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// stack_push_float(Queue:me,Float:newval)
static cell AMX_NATIVE_CALL	stack_push_float(AMX *amx,cell *params)
{
	Capsule tmp; tmp.setfloat(amx_ctof(params[2]));
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// stack_push_stack(Queue:me,Float:newval[3])
static cell AMX_NATIVE_CALL	stack_push_stack(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	HLVector tty(amx_ctof(ptr[0]),amx_ctof(ptr[1]),amx_ctof(ptr[2]));

	Capsule tmp; tmp.setvector(tty);
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// stack_push_array(Queue:me,newval[],len)
static cell AMX_NATIVE_CALL	stack_push_array(AMX *amx,cell *params)
{
	cell* ptr = MF_GetAmxAddr(amx,params[2]);
	vector<cell> tty(ptr,params[3]);

	Capsule tmp; tmp.setarray(tty);
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// stack_push_string(Queue:me,newval[])
static cell AMX_NATIVE_CALL	stack_push_string(AMX *amx,cell *params)
{
	int iLen;
	char* ptr = MF_GetAmxString(amx,params[2],0,&iLen);
	String tty(ptr,iLen);

	Capsule tmp; tmp.setstring(tty);
	reinterpret_cast<any_stack*>(params[1])->push(tmp);
	return 1;
}

// Any:stack_peek(Queue:me)
static cell AMX_NATIVE_CALL	stack_peek(AMX *amx,cell *params)
{
	Capsule& tmp = reinterpret_cast<any_stack*>(params[1])->peek();
	return reinterpret_cast<cell>(&tmp);
}

// stack_peek_raw(Queue:me)
static cell AMX_NATIVE_CALL	stack_peek_raw(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	
	return reinterpret_cast<cell>(tmp_stack->peek().getData() );
}

// bool:stack_peek_bool(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_peek_bool(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	bool* tmp = tmp_stack->peek().getbool();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve boolean from non-boolean ANY class"); }
		return -1;
	}
	return static_cast<cell>(*tmp);
}

// cell:stack_peek_int(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_peek_int(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	cell* tmp = tmp_stack->peek().getint();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve integer from non-integer ANY class"); }
		return -1;
	}
	return *tmp;
}

// Ptr:stack_peek_ptr(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_peek_ptr(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	void* tmp = tmp_stack->peek().getpointer();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve pointer from non-pointer ANY class"); }
		return -1;
	}
	return reinterpret_cast<cell>(tmp);
}

// Float:stack_peek_float(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_peek_float(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	REAL* tmp = tmp_stack->peek().getfloat();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve float from non-float ANY class"); }
		return -1;
	}
	return amx_ftoc(*tmp);
}

// stack_peek_stack(Queue:me,Float:vec[3],no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_peek_stack(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	HLVector* tmp = tmp_stack->peek().getvector();

	if( (tmp == NULL) )
	{
		if(params[3] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve stack from non-stack ANY class"); }
		return -1;
	}
	cell* vec = MF_GetAmxAddr(amx,params[2]);

	vec[0] = amx_ftoc(tmp->x);
	vec[1] = amx_ftoc(tmp->y);
	vec[2] = amx_ftoc(tmp->z);

	return 1;
}

// stack_peek_array(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_peek_array(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	register vector<cell>* tmp = tmp_stack->peek().getarray();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve array from non-array ANY class"); }
		return -1;
	}
	register cell* arr = MF_GetAmxAddr(amx,params[2]);
	register size_t len = (static_cast<ucell>(params[3]) < tmp->size()) ? (params[3]) : (tmp->size());
	for(register size_t i = 0; i < len; i++)
	{
		arr[i] = (*tmp)[i];
	}

	return len;
}

// stack_peek_string(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_peek_string(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	register String* tmp = tmp_stack->peek().getstring();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve string from non-string ANY class"); }
		return -1;
	}

	return MF_SetAmxString(amx,params[2],tmp->c_str(),params[3]);
}

// Any:stack_top(Queue:me)
static cell AMX_NATIVE_CALL	stack_top(AMX *amx,cell *params)
{
	return reinterpret_cast<cell>(&reinterpret_cast<any_stack*>(params[1])->top() );
}

// stack_top_raw(Queue:me)
static cell AMX_NATIVE_CALL	stack_top_raw(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	
	return reinterpret_cast<cell>(tmp_stack->top().getData() );
}

// bool:stack_top_bool(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_top_bool(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	bool* tmp = tmp_stack->top().getbool();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve boolean from non-boolean ANY class"); }
		return -1;
	}
	return static_cast<cell>(*tmp);
}

// cell:stack_top_int(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_top_int(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	cell* tmp = tmp_stack->top().getint();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve integer from non-integer ANY class"); }
		return -1;
	}
	return *tmp;
}

// Ptr:stack_top_ptr(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_top_ptr(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	void* tmp = tmp_stack->top().getpointer();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve pointer from non-pointer ANY class"); }
		return -1;
	}
	return reinterpret_cast<cell>(tmp);
}

// Float:stack_top_float(Queue:me,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_top_float(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	REAL* tmp = tmp_stack->top().getfloat();

	if( (tmp == NULL) )
	{
		if(params[2] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve float from non-float ANY class"); }
		return -1;
	}
	return amx_ftoc(*tmp);
}

// stack_top_stack(Queue:me,Float:vec[3],no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_top_stack(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	HLVector* tmp = tmp_stack->top().getvector();

	if( (tmp == NULL) )
	{
		if(params[3] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve stack from non-stack ANY class"); }
		return -1;
	}
	cell* vec = MF_GetAmxAddr(amx,params[2]);

	vec[0] = amx_ftoc(tmp->x);
	vec[1] = amx_ftoc(tmp->y);
	vec[2] = amx_ftoc(tmp->z);

	return 1;
}

// stack_top_array(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_top_array(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	register vector<cell>* tmp = tmp_stack->top().getarray();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve array from non-array ANY class"); }
		return -1;
	}
	register cell* arr = MF_GetAmxAddr(amx,params[2]);
	register size_t len = (static_cast<ucell>(params[3]) < tmp->size()) ? (params[3]) : (tmp->size());
	for(register size_t i = 0; i < len; i++)
	{
		arr[i] = (*tmp)[i];
	}

	return len;
}

// stack_top_string(Queue:me,arr[],len,no_fatal_errors=0)
static cell AMX_NATIVE_CALL	stack_top_string(AMX *amx,cell *params)
{
	any_stack* tmp_stack = reinterpret_cast<any_stack*>(params[1]);
	register String* tmp = tmp_stack->top().getstring();

	if( (tmp == NULL) )
	{
		if(params[4] == 0) { MF_LogError(amx,AMX_ERR_NATIVE,"Attempted to retrieve string from non-string ANY class"); }
		return -1;
	}

	return MF_SetAmxString(amx,params[2],tmp->c_str(),params[3]);
}

AMX_NATIVE_INFO stack_Natives[] =
{
	{"stack_create",		stack_create},
	{"stack_destroy",		stack_destroy},

	{"stack_clear",		stack_clear},
	{"stack_copy",		stack_copy},

	{"stack_comp",		stack_comp},
	{"stack_lex",		stack_lex},

	{"stack_empty",		stack_empty},
	{"stack_size",		stack_size},
	{"stack_memory",	stack_memory},

	{"stack_pop",		stack_pop},

	{"stack_peek",				stack_peek},
	{"stack_peek_raw",			stack_peek_raw},
	{"stack_peek_bool",			stack_peek_bool},
	{"stack_peek_int",			stack_peek_int},
	{"stack_peek_float",		stack_peek_float},
	{"stack_peek_ptr",			stack_peek_ptr},
	{"stack_peek_stack",		stack_peek_stack},
	{"stack_peek_array",		stack_peek_array},
	{"stack_peek_string",		stack_peek_string},

	{"stack_push",				stack_push},
	{"stack_push_none",			stack_push_none},
	{"stack_push_bool",			stack_push_bool},
	{"stack_push_int",			stack_push_int},
	{"stack_push_float",		stack_push_float},
	{"stack_push_ptr",			stack_push_ptr},
	{"stack_push_stack",		stack_push_stack},
	{"stack_push_array",		stack_push_array},
	{"stack_push_string",		stack_push_string},

	{"stack_top",				stack_top},
	{"stack_top_raw",			stack_top_raw},
	{"stack_top_bool",		stack_top_bool},
	{"stack_top_int",			stack_top_int},
	{"stack_top_float",		stack_top_float},
	{"stack_top_ptr",			stack_top_ptr},
	{"stack_top_stack",		stack_top_stack},
	{"stack_top_array",		stack_top_array},
	{"stack_top_string",		stack_top_string},

	{ NULL,				NULL}
};